(* weak normalization for typable combinator terms (expression style) *)
(*** Uses a pattern-matching definition of Computability ***)
Reset Initial.

(*  Object level  *)
Parameter AP:Set.    (* a type of atomic propositions (types) *)

(* object-level propositions (types). *)
Inductive prop: Set := atom:(a:AP)prop | arrow:prop->prop->prop.

(* infix operator for "arrow".  All Coq infixes are right associative *)
Infix RIGHTA 6 ">>" arrow.

(* object-level terms. *)
Inductive term:Set := k:term | s:term | app:term->term->term.

(* infix operator for "app" *)
Infix 5 "^" app.

(* object-level provability judgement. *)
Inductive thm:term->prop->Prop :=
   K:(p,q:prop)(thm k p>>q>>p)
 | S:(p,q,r:prop)(thm s (p>>q>>r) >> (p>>q) >> p>>r)
 | MP:(a,b:term)(p,q:prop)
        (lp:(thm a p>>q))   (rp:(thm b p))
     (************************************)
              (thm a^b q).

(* unique construction *)
Lemma k_not_s: ~(k=s).
Unfold not. Intro kqs. Discriminate kqs.
Save.

(* constructors are injections *)
Lemma arr_inj: (p,q,p',q':prop)(p>>q) = (p'>>q') -> p=p' /\ q=q'.
Intros. Injection H. Intros. Auto.
Save.

(* inversions *)
Lemma invertK:
  (p:prop)(thm k p)->(EX r:prop | (EX t:prop | p = (r>>t>>r))).
Intros. Inversion H. Exists p0. Exists q. Reflexivity.
Save.

Lemma invertMP:
  (M,N:term)(p:prop)(thm M^N p)->
      (EX q:prop | (thm M q>>p) /\ (thm N q)).
Intros. Inversion H. Exists p0. Auto.
Save.

(* Define the reduction relation on object-level terms. *)
Inductive red:term->term->Prop :=
    kred:(a,b:term)(red k^a^b a)
  | sred:(a,b,c:term)(red  s^a^b^c  a^c^(b^c))
  | l_cong:(a,a':term)(b:term)
            (red a a')->
        (**************************)
            (red a^b a'^b)
  | r_cong:(a,a':term)(b:term)
            (red a a')->
        (**************************)
            (red b^a b^a').

Tactic Definition rewritesThenAssumption :=
Match Context With
  [ id : ? = ? |- ? ] ->
       Rewrite id; Assumption
| [ id : ? = ? |- ? ] -> Rewrite <- id;Assumption
| [ id : ? = ? |- ? ] ->
	Rewrite id;Clear id; rewritesThenAssumption
| [ id : ? = ? |- ? ] ->
	Rewrite <- id;Clear id; rewritesThenAssumption.


(* subject reduction *)
Lemma subjRed: (M,N:term)(red M N)->(a:prop)(thm M a)->(thm N a).
Induction 1; Intros.
(* kred *)
Inversion H0. Inversion lp. Inversion lp0. Rewrite <- H11. Assumption.
(* sred *)
Inversion H0. Inversion lp.  Inversion lp0. Inversion lp1.
Apply (MP (a^c) (b^c) q2).  Apply (MP a c p2).
rewritesThenAssumption.
rewritesThenAssumption.

Apply (MP b c p2).
rewritesThenAssumption.
rewritesThenAssumption.

(* l_cong *)
Inversion H2. Apply (MP a' b p); Auto.
(* r_cong *)
Inversion H2. Apply (MP b a' p); Auto.
   (**  exercise  **)
Save.

(* Normalizing *)
Inductive WNorm:term->Prop :=
    ntk:(WNorm k)
  | nts:(WNorm s)
  | ntkM:(M:term)
           (Mn:(WNorm M))
        (**************************)
           (WNorm k^M)
  | ntsM:(M:term)
           (Mn:(WNorm M))
        (**************************)
           (WNorm s^M)
  | ntsMN:(M,N:term)
           (Mn:(WNorm M))
           (Nn:(WNorm N))
        (**************************)
           (WNorm s^M^N)
  | ntStep:(M,N:term)(p:(red M N))
           (Nn:(WNorm N))
        (**************************)
           (WNorm M).


(* Define computability by induction on prop:
 * No term of atomic type is computable (without object level 
 *   assumptions there is no well-typed term of atomic type)
 * A term of type x1>>x2 is computable if it is Normalizing, and if
 *   for all terms N of type x1, if N is computable then (M*N) is
 *   computable.
 *)
Definition Comp : (a:prop)(M:term)Prop.
Induction 1.
Intros; Exact False.
Intros. Apply (and (WNorm M) ((N:term)(X N)->(X0 M^N))).
Defined.


(* some properties of computability *)
Lemma CompWNorm: (a:prop)(M:term)(Comp a M)->(WNorm M).
Induction a.
Intros. Elim H.
Intros. Exact (proj1 ?? H1).
Save.

Lemma appPreserveComp:
      (a,b:prop)(M,N:term)(Comp a>>b M)->(Comp a N)->(Comp b M^N).
Intros. Exact (proj2 ?? H ? H0).
Save.

Lemma ExpandPreserveComp:
     (a:prop)(M,N:term)(red M N)->(Comp a N)->(Comp a M).
Induction a.
Intros. Elim H0.
Intros. Split.
  Exact (ntStep ?? H1 (CompWNorm ?? H2)).
  Intros O X. Apply (H0 ?? (l_cong ?? O H1)).
  Apply (appPreserveComp ???? H2 X).
Save.

Lemma kComp: (p,q:prop)(Comp p>>q>>p k).
Intros. Split. Exact ntk.
Intros. Split. Apply ntkM. Exact (CompWNorm ?? H).
Intros M HM. Apply (ExpandPreserveComp p k^N^M ? (kred N M)). Auto.
Save.

Lemma sComp : (p,q,r:prop)(Comp ((p>>(q>>r))>>((p>>q)>>p>>r)) s).
Intros. Split. Exact nts.
Intros. Split. Apply ntsM.  Apply (CompWNorm ? ? H).
Intros. Split. Apply ntsMN. Apply (CompWNorm ? ? H).
Apply (CompWNorm ? ? H0).
Intros. Apply (ExpandPreserveComp r ? ? (sred N N0 N1)).
Apply appPreserveComp with q;Auto.
Apply appPreserveComp with p;Auto.
Apply appPreserveComp with p;Auto.
Save.

(** main theorem **)

Lemma AllComp : (M:term;a:prop)(thm M a)->(Comp a M).
Induction 1.
(* k *)
Apply kComp.
(* s *)
Apply sComp.
(* app *)
Intros. Apply (appPreserveComp p);Assumption .
Save.

(* all terms are normalizing *)
Theorem AllWNorm: (a:prop; M:term)(thm M a)->(WNorm M).
Intros; EApply CompWNorm;EApply AllComp;EAuto.
Save.


