(*Object level  *)
Parameter AP : Set.
(* a type of atomic propositions (types) 
   object-level propositions (types). *)
 
Inductive prop : Set :=
  atom: (a : AP)  prop
 | arrow: prop -> prop ->  prop .
(* object-level terms. *)
 
Inductive term : Set :=
  k: term
 | s: term
 | app: term -> term ->  term .
(* object-level provability judgement. 
  **********************************
   object-level provability judgement. 
  ***********************************)
 
Inductive thm : term -> prop ->  Prop :=
  K: (p, q : prop)  (thm k (arrow p (arrow q p)))
 | S:
     (p, q, r : prop)
      (thm s (arrow (arrow p (arrow q r)) (arrow (arrow p q) (arrow p r))))
 | MP:
     (a, b : term)
     (p, q : prop)
     (lp : (thm a (arrow p q))) (rp : (thm b p))  (thm (app a b) q) .
(* unique construction *)
 
Lemma k_not_s: ~ k = s.
Unfold not.
Intros kqs.
Discriminate kqs.
Qed.
(* constructors are injections *)
 
Lemma arr_inj:
 (p, q, p', q' : prop) (arrow p q) = (arrow p' q') ->  p = p' /\ q = q'.
Intros.
Injection H.
Intros.
Auto.
Qed.
(* inversions *)
 
Lemma invertK:
 (p : prop)
 (thm k p) ->
  (ex prop [r : prop]  (ex prop [t : prop]  p = (arrow r (arrow t r)))).
Intros.
Inversion H.
Exists p0.
Exists q.
Reflexivity.
Qed.
 
Lemma invertMP:
 (M, N : term)
 (p : prop)
 (thm (app M N) p) ->  (ex prop [q : prop]  (thm M (arrow q p)) /\ (thm N q)).
Intros.
Inversion H.
Exists p0.
Auto.
Qed.
(* Define the reduction relation on object-level terms. 
  ************************
  *************************)
 
Inductive red : term -> term ->  Prop :=
  kred: (a, b : term)  (red (app (app k a) b) a)
 | sred:
     (a, b, c : term)  (red (app (app (app s a) b) c) (app (app a c) (app b c)))
 | l_cong: (a, a' : term) (b : term) (red a a') ->  (red (app a b) (app a' b))
 | r_cong: (a, a' : term) (b : term) (red a a') ->  (red (app b a) (app b a')) .
 
Tactic Definition rewritesThenAssumption :=
 (Match Context With
 | [id:? = ? |- ?] -> Rewrite id; Assumption
 | [id:? = ? |- ?] -> Rewrite <- id; Assumption
 | [id:? = ? |- ?] -> Rewrite id; Clear id; rewritesThenAssumption
 | [id:? = ? |- ?] -> Rewrite <- id; Clear id; rewritesThenAssumption).
(* subject reduction *)
 
Lemma subjRed: (M, N : term) (red M N) -> (a : prop) (thm M a) ->  (thm N a).
Induction 1; Intros.
(* kred *)
Inversion H0.
Inversion lp.
Inversion lp0.
Rewrite <- H11.
Assumption.
(* sred *)
Inversion H0.
Inversion lp.
Inversion lp0.
Inversion lp1.
Apply (MP (app a c) (app b c) q2).
Apply (MP a c p2).
rewritesThenAssumption.
rewritesThenAssumption.
Apply (MP b c p2).
rewritesThenAssumption.
rewritesThenAssumption.
(* l_cong *)
Inversion H2.
Apply (MP a' b p); Auto.
(* r_cong *)
Inversion H2.
Apply (MP b a' p); Auto.
(**  exercise  **)
Qed.
(* Normalizing 
  ************************
  ************************
  ************************
  *************************)
 
Inductive WNorm : term ->  Prop :=
  ntk: (WNorm k)
 | nts: (WNorm s)
 | ntkM: (M : term) (Mn : (WNorm M))  (WNorm (app k M))
 | ntsM: (M : term) (Mn : (WNorm M))  (WNorm (app s M))
 | ntsMN:
     (M, N : term) (Mn : (WNorm M)) (Nn : (WNorm N))  (WNorm (app (app s M) N))
 | ntStep: (M, N : term) (p : (red M N)) (Nn : (WNorm N))  (WNorm M) .
(* Define computability by induction on prop:
   * No term of atomic type is computable (without object level 
   *   assumptions there is no well-typed term of atomic type)
   * A term of type x1>>x2 is computable if it is Normalizing, and if
   *   for all terms N of type x1, if N is computable then (M*N) is
   *   computable.
   *)
 
Definition Comp: (a : prop) (M : term)  Prop.
Induction 1.
Intros; Exact False.
Intros.
Apply (WNorm M) /\ ((N : term) (X N) ->  (X0 (app M N))).
Defined.
(* some properties of computability *)
 
Lemma CompWNorm: (a : prop) (M : term) (Comp a M) ->  (WNorm M).
Induction a.
Intros.
Elim H.
Intros.
Exact <?, ?> Fst {H1}.
Qed.
 
Lemma appPreserveComp:
 (a, b : prop)
 (M, N : term) (Comp (arrow a b) M) -> (Comp a N) ->  (Comp b (app M N)).
Intros.
Exact (proj2 ? ? H ? H0).
Qed.
 
Lemma ExpandPreserveComp:
 (a : prop) (M, N : term) (red M N) -> (Comp a N) ->  (Comp a M).
Induction a.
Intros.
Elim H0.
Intros.
Split.
Exact (ntStep ? ? H1 (CompWNorm ? ? H2)).
Intros O X.
Apply (H0 ? ? (l_cong ? ? O H1)).
Apply (appPreserveComp ? ? ? ? H2 X).
Qed.
 
Lemma kComp: (p, q : prop)  (Comp (arrow p (arrow q p)) k).
Intros.
Split.
Exact ntk.
Intros.
Split.
Apply ntkM.
Exact (CompWNorm ? ? H).
Intros M HM.
Apply (ExpandPreserveComp p (app (app k N) M) ? (kred N M)).
Auto.
Qed.
 
Lemma sComp:
 (p, q, r : prop)
  (Comp (arrow (arrow p (arrow q r)) (arrow (arrow p q) (arrow p r))) s).
Intros.
Split.
Exact nts.
Intros.
Split.
Apply ntsM.
Apply (CompWNorm ? ? H).
Intros.
Split.
Apply ntsMN.
Apply (CompWNorm ? ? H).
Apply (CompWNorm ? ? H0).
Intros.
Apply (ExpandPreserveComp r ? ? (sred N N0 N1)).
Apply appPreserveComp with q; Auto.
Apply appPreserveComp with p; Auto.
Apply appPreserveComp with p; Auto.
Qed.
(** main theorem **)
 
Lemma AllComp: (M : term) (a : prop) (thm M a) ->  (Comp a M).
Induction 1.
(* k *)
Apply kComp.
(* s *)
Apply sComp.
(* app *)
Intros.
Apply (appPreserveComp p); Assumption.
Qed.
(* all terms are normalizing *)
 
Theorem AllWNorm: (a : prop) (M : term) (thm M a) ->  (WNorm M).
Intros; EApply CompWNorm; EApply AllComp; EAuto.
Qed.
